@@ -505,7 +505,7 @@ setOccursCheck(atom_t a)
 static int
 setEncoding(atom_t a)
 { GET_LD
-  IOENC enc = atom_to_encoding(a);
+  IOENC enc = PL_atom_to_encoding(a);
 
   if ( enc == ENC_UNKNOWN )
   { term_t value = PL_new_term_ref();
@@ -810,15 +810,11 @@ set_prolog_flag_unlocked(DECL_LD Module m, atom_t k, term_t value, int flags)
       { debugstatus.showContext = val;
 #ifdef O_PLMT
       } else if ( k == ATOM_threads )
-      { if ( val )
+      { if ( !!val != !!GD->thread.enabled )
 	{ rval = enableThreads(val);
-	  PL_LOCK(L_PLFLAG);
-	} else
-	{ PL_UNLOCK(L_PLFLAG);
-	  rval = enableThreads(val);
+	  if ( !rval )
+	    break;			/* do not change value */
 	}
-	if ( !rval )
-	  break;			/* don't change value */
 #endif
       } else if ( k == ATOM_tty_control )
       { if ( val != (f->value.a == ATOM_true) )
@@ -970,6 +966,8 @@ set_prolog_flag(term_t key, term_t value, int flags)
 
   if ( k == ATOM_autoload && !propagateAutoload(value) )
     return FALSE;
+  if ( k == ATOM_threads )
+    return set_prolog_flag_unlocked(m, k, value, flags);
 
   PL_LOCK(L_PLFLAG);
   rc = set_prolog_flag_unlocked(m, k, value, flags);
@@ -1388,6 +1386,37 @@ pl_prolog_flag(term_t name, term_t value, control_t h)
 }
 
 
+static void
+set_arch(void)
+{
+#ifdef __APPLE__
+#include <sys/sysctl.h>
+  uint32_t cputype;
+  size_t size = sizeof(cputype);
+
+  if ( strcmp(PLARCH, "fat-darwin") == 0 )
+  { setPrologFlag("apple_universal_binary", FT_BOOL|FF_READONLY, TRUE, 0);
+
+    if ( sysctlbyname("hw.cputype", &cputype, &size, NULL, 0) == 0 )
+    { switch(cputype&0xff)
+      { case 7:
+	  setPrologFlag("arch", FT_ATOM|FF_READONLY, "x86_64-darwin");
+          return;
+        case 12:
+	  setPrologFlag("arch", FT_ATOM|FF_READONLY, "arm64-darwin");
+          return;
+        default:
+	  Sdprintf("sysctlbyname() cputype = %d (unknown)\n", (int)cputype);
+      }
+    }
+  }
+#endif
+
+  setPrologFlag("arch", FT_ATOM|FF_READONLY, PLARCH);
+}
+
+
+
 		 /*******************************
 		 *	INITIALISE FEATURES	*
 		 *******************************/
@@ -1407,7 +1436,7 @@ initPrologFlagTable(void)
 { if ( !GD->prolog_flag.table )
   { initPrologThreads();	/* may be called before PL_initialise() */
 
-    GD->prolog_flag.table = newHTable(64);
+    GD->prolog_flag.table = newHTable(256);
   }
 }
 
@@ -1415,8 +1444,9 @@ initPrologFlagTable(void)
 void
 initPrologFlags(void)
 { GET_LD
+
   setPrologFlag("iso",  FT_BOOL, FALSE, PLFLAG_ISO);
-  setPrologFlag("arch", FT_ATOM|FF_READONLY, PLARCH);
+  set_arch();
 #if __WINDOWS__
   setPrologFlag("windows",	FT_BOOL|FF_READONLY, TRUE, 0);
   const char *wine_version;
@@ -1603,7 +1633,8 @@ initPrologFlags(void)
 #endif
 #endif
 
-  setPrologFlag("encoding", FT_ATOM, stringAtom(encoding_to_atom(LD->encoding)));
+  setPrologFlag("encoding", FT_ATOM,
+		stringAtom(PL_encoding_to_atom(LD->encoding)));
 
   setPrologFlag("tty_control", FT_BOOL,
 		truePrologFlag(PLFLAG_TTY_CONTROL), PLFLAG_TTY_CONTROL);
@@ -1684,7 +1715,7 @@ static void
 setTZPrologFlag(void)
 { tzset();
 
-#ifdef _MSC_VER
+#if defined(__WINDOWS__) || defined(__CYGWIN__)
 #define timezone _timezone
 #endif
 
