@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2020, University of Amsterdam
-                         VU University Amsterdam
-		         CWI, Amsterdam
+    Copyright (c)  2020-2022, University of Amsterdam
+			      VU University Amsterdam
+			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -196,13 +197,22 @@ int
 transaction_visible_clause(DECL_LD Clause cl, gen_t gen)
 { if ( cl->generation.created <= LD->transaction.gen_start &&
        cl->generation.erased   > LD->transaction.gen_start )
-  { intptr_t lgen;
+  { if ( cl->tr_erased_no )
+    { tr_stack s = { .clauses = LD->transaction.clauses,
+                     .parent  = LD->transaction.stack
+		   };
+      tr_stack *stack;
 
-    if ( cl->tr_erased_no && LD->transaction.clauses &&
-	 (lgen = (intptr_t)lookupHTable(LD->transaction.clauses, cl)) &&
-	 !IS_ASSERT_GEN(lgen) )
-    { if ( lgen+LD->transaction.gen_base <= gen )
-	return FALSE;
+      for(stack=&s; stack; stack=stack->parent)
+      { intptr_t lgen;
+
+	if ( stack->clauses &&
+	     (lgen = (intptr_t)lookupHTable(stack->clauses, cl)) &&
+	     !IS_ASSERT_GEN(lgen) )
+	{ if ( lgen+LD->transaction.gen_base <= gen )
+	    return FALSE;
+	}
+      }
     }
 
     return TRUE;
@@ -291,36 +301,36 @@ transaction_commit(DECL_LD)
     PL_LOCK(L_GENERATION);
     gen_commit = global_generation()+1;
 
-    for_table(LD->transaction.clauses, n, v,
-	      { Clause cl = n;
-		uintptr_t lgen = (uintptr_t)v;
+    FOR_TABLE(LD->transaction.clauses, n, v)
+    { Clause cl = n;
+      uintptr_t lgen = (uintptr_t)v;
 
-		if ( IS_ASSERT_GEN(lgen) )
-		{ if ( false(cl, CL_ERASED) )
-		  { cl->generation.erased  = GEN_MAX;
-		    MEMORY_RELEASE();
-		    cl->generation.created = gen_commit;
-		    DEBUG(MSG_COMMIT,
-			Sdprintf("Commit added clause %p for %s\n",
-				 cl, predicateName(cl->predicate)));
-		  } else
-		  { DEBUG(MSG_COMMIT,
-			  Sdprintf("Discarded in-transaction clause %p for %s\n",
-				   cl, predicateName(cl->predicate)));
-		    cl->generation.erased  = GEN_TR_ASSERT_ERASE;
-		    MEMORY_RELEASE();
-		    cl->generation.created = GEN_TR_ASSERT_ERASE;
-		  }
-		} else if ( lgen == GEN_NESTED_RETRACT )
-		{ retract_clause(cl, gen_commit);
-		} else
-		{ DEBUG(MSG_COMMIT,
-			Sdprintf("Commit erased clause %p for %s\n",
-				 cl, predicateName(cl->predicate)));
-		  ATOMIC_DEC(&cl->tr_erased_no);
-		  retract_clause(cl, gen_commit);
-		}
-	      });
+      if ( IS_ASSERT_GEN(lgen) )
+      { if ( false(cl, CL_ERASED) )
+	{ cl->generation.erased  = GEN_MAX;
+	  MEMORY_RELEASE();
+	  cl->generation.created = gen_commit;
+	  DEBUG(MSG_COMMIT,
+	      Sdprintf("Commit added clause %p for %s\n",
+		       cl, predicateName(cl->predicate)));
+	} else
+	{ DEBUG(MSG_COMMIT,
+		Sdprintf("Discarded in-transaction clause %p for %s\n",
+			 cl, predicateName(cl->predicate)));
+	  cl->generation.erased  = GEN_TR_ASSERT_ERASE;
+	  MEMORY_RELEASE();
+	  cl->generation.created = GEN_TR_ASSERT_ERASE;
+	}
+      } else if ( lgen == GEN_NESTED_RETRACT )
+      { retract_clause(cl, gen_commit);
+      } else
+      { DEBUG(MSG_COMMIT,
+	      Sdprintf("Commit erased clause %p for %s\n",
+		       cl, predicateName(cl->predicate)));
+	ATOMIC_DEC(&cl->tr_erased_no);
+	retract_clause(cl, gen_commit);
+      }
+    }
     MEMORY_RELEASE();
     GD->_generation = gen_commit;
     PL_UNLOCK(L_GENERATION);
@@ -391,28 +401,29 @@ transaction_discard(DECL_LD)
 }
 
 static void
-merge_tables(Table into, Table from)
-{ for_table(from, n, v,
-	    { Clause cl = n;
+merge_clause_tables(Table into, Table from)
+{ FOR_TABLE(from, n, v)
+  { Clause cl = n;
 
-	      acquire_clause(cl);
-	      addHTable(into, n, v);
-	    });
+    if ( addHTable(into, cl, v) == v )
+      acquire_clause(cl);		/* new in outer table */
+    else
+      updateHTable(into, cl, v);	/* already in outer table */
+  };
 }
 
 #define merge_pred_tables(into, from) LDFUNC(merge_pred_tables, into, from)
 static void
 merge_pred_tables(DECL_LD Table into, Table from)
-{ for_table(from, n, v,
-	    { Definition def = n;
-	      void *lgen = v;
-	      void *lgen0 = lookupHTable(into, def);
-	      int oflags = PTR_GEN_FLAGS(lgen0);
+{ FOR_TABLE(from, n, v)
+  { Definition def = n;
+    void *lgen = v;
+    void *lgen0 = lookupHTable(into, def);
+    int oflags = PTR_GEN_FLAGS(lgen0);
 
 
-	      updateHTable(into, def,
-			   PTR_ADD_FLAGS(lgen, oflags));
-	    });
+    updateHTable(into, def, PTR_ADD_FLAGS(lgen, oflags));
+  };
 }
 
 
@@ -556,15 +567,13 @@ transaction(DECL_LD term_t goal, term_t constraint, term_t lock, int flags)
       }
 
       if ( LD->transaction.clauses )
-      {
-
-	if ( parent.clauses )
+      {	if ( parent.clauses )
 	{ if ( (flags&TR_BULK) )
 	  { transaction_updates(&updates);
 	    if ( !announce_updates(&updates) )
 	      goto nested_discard;
 	  }
-	  merge_tables(parent.clauses, LD->transaction.clauses);
+	  merge_clause_tables(parent.clauses, LD->transaction.clauses);
 	  destroyHTable(LD->transaction.clauses);
 	} else
 	{ parent.clauses = LD->transaction.clauses;
@@ -618,11 +627,13 @@ transaction(DECL_LD term_t goal, term_t constraint, term_t lock, int flags)
 	if ( constraint ) TR_UNLOCK();
       } else
       { if ( constraint ) TR_UNLOCK();
+	LD->transaction.generation = 0;	/* avoid recording the rollback */
 	transaction_discard();
 	transaction_rollback_tables();
       }
     } else
-    { rc = transaction_discard() && rc;
+    { LD->transaction.generation = 0;  /* avoid recording the rollback */
+      rc = transaction_discard() && rc;
       rc = transaction_rollback_tables() && rc;
     }
     LD->transaction.id         = 0;
@@ -656,7 +667,7 @@ PRED_IMPL("$transaction", 2, transaction, PL_FA_TRANSPARENT)
   if ( bulk )
     flags |= TR_BULK;
 
-  return transaction(A1, 0, 0, TR_TRANSACTION);
+  return transaction(A1, 0, 0, flags);
 }
 
 static
