@@ -316,12 +316,10 @@ updateAlerted(PL_local_data_t *ld)
 
 int
 raiseSignal(PL_local_data_t *ld, int sig)
-{ if ( sig > 0 && sig <= MAXSIGNAL && ld )
-  { int off = (sig-1) / 32;
-    int mask = (1 << ((sig-1)%32));
-    int alerted;
+{ if ( IS_VALID_SIGNAL(sig) && ld )
+  { int alerted;
 
-    ATOMIC_OR(&ld->signal.pending[off], mask);
+    WSIGMASK_SET(ld->signal.pending, sig);
 
     do
     { alerted = ld->alerted;
@@ -336,11 +334,8 @@ raiseSignal(PL_local_data_t *ld, int sig)
 
 int
 pendingSignal(PL_local_data_t *ld, int sig)
-{ if ( sig > 0 && sig <= MAXSIGNAL && ld )
-  { int off  = (sig-1)/32;
-    int mask = 1 << ((sig-1)%32);
-
-    return (ld->signal.pending[off] & mask) ? TRUE : FALSE;
+{ if ( IS_VALID_SIGNAL(sig) && ld )
+  { return WSIGMASK_ISSET(ld->signal.pending, sig) ? TRUE : FALSE;
   }
 
   return -1;
@@ -495,13 +490,23 @@ PL_discard_foreign_frame(fid_t id)
 }
 
 
-#define determinism_error(fr, found) LDFUNC(determinism_error, fr, found)
+#define determinism_error(fr, bfr, found) \
+	LDFUNC(determinism_error, fr, bfr, found)
+
 static int
-determinism_error(DECL_LD LocalFrame fr, atom_t found)
+determinism_error(DECL_LD LocalFrame fr, Choice bfr, atom_t found)
 { fid_t fid;
   int rc = FALSE;
   atom_t a = ATOM_error;
 
+  if ( found == ATOM_nondet )
+  { for(; bfr && bfr->type == CHP_DEBUG; bfr=bfr->parent)
+      ;
+
+    if ( (void*)bfr < (void*)fr )
+      return TRUE;
+  }
+
   PL_current_prolog_flag(ATOM_determinism_error, PL_ATOM, &a);
   if ( a == ATOM_silent )
     return TRUE;
@@ -573,7 +578,7 @@ ssu_or_det_failed(DECL_LD LocalFrame fr)
   int rc = FALSE;
 
   if ( false(fr, FR_SSU_DET) )
-    return determinism_error(fr, ATOM_fail);
+    return determinism_error(fr, NULL, ATOM_fail);
 
   if ( (fid = PL_open_foreign_frame()) )
   { term_t goal;
@@ -827,11 +832,7 @@ callCleanupHandler(DECL_LD LocalFrame fr, enum finished reason)
       fr = (LocalFrame)valTermRef(fref);
       clean = consTermRef(argFrameP(fr, 3));
       if ( saveWakeup(&wstate, FALSE) )
-      { static predicate_t PRED_call1 = NULL;
-	int rval;
-
-	if ( !PRED_call1 )
-	  PRED_call1 = PL_predicate("call", 1, "system");
+      { int rval;
 
 	startCritical;
         rval = call_term(contextModule(fr), clean);
@@ -2111,7 +2112,7 @@ exception_hook(DECL_LD qid_t pqid, term_t fr, term_t catchfr_ref)
       PL_put_frame(av+2, (LocalFrame)valTermRef(fr));
 
       if ( !catchfr_ref )
-	catchfr_ref = isCaughtInOuterQuery(pqid, exception_term);
+	catchfr_ref = isCaughtInOuterQuery(pqid, exception_bin);
       if ( catchfr_ref == (term_t)-1 )
       { PL_put_atom_chars(av+3, "C");
       } else if ( catchfr_ref && catchfr_ref == LD->exception.fr_rewritten )
@@ -2129,9 +2130,11 @@ exception_hook(DECL_LD qid_t pqid, term_t fr, term_t catchfr_ref)
       { PL_put_frame(av+3, NULL);	/* puts 'none' */
       }
 
+      startCritical;
       qid = PL_open_query(MODULE_user, PL_Q_NODEBUG|PL_Q_CATCH_EXCEPTION,
 			  PROCEDURE_exception_hook4, av);
       rc = PL_next_solution(qid);
+      rc = endCritical && rc;
       debug = debugstatus.debugging;
       trace = debugstatus.tracing;
       if ( rc )				/* pass user setting trace/debug */
@@ -2696,26 +2699,30 @@ PL_open_query(Module ctx, int flags, Procedure proc, term_t args)
   environment_frame = fr;
   qf->parent = LD->query;
   LD->query = qf;
+  qf->qid = allocHeapOrHalt(sizeof(*qf->qid));
+  qf->qid->engine = LD;
+  qf->qid->offset = consTermRef(qf);
 
-  DEBUG(2, Sdprintf("QID=%d\n", QidFromQuery(qf)));
+  DEBUG(2, Sdprintf("QID=%p\n", QidFromQuery(qf)));
   updateAlerted(LD);
 
   return QidFromQuery(qf);
 }
 
 
-#define discard_query(qid) LDFUNC(discard_query, qid)
 static void
-discard_query(DECL_LD qid_t qid)
+discard_query(qid_t qid)
 { QueryFrame qf = QueryFromQid(qid);
 
-  discardChoicesAfter(&qf->frame, FINISH_CUT);
-  qf = QueryFromQid(qid);		/* may be shifted */
-  discardFrame(&qf->frame);
-  if ( true(&qf->frame, FR_WATCHED) )
-  { lTop = (LocalFrame)argFrameP(&qf->frame,
-				 qf->frame.predicate->functor->arity);
-    frameFinished(&qf->frame, FINISH_CUT);
+  WITH_LD(qid->engine)
+  { discardChoicesAfter(&qf->frame, FINISH_CUT);
+    qf = QueryFromQid(qid);		/* may be shifted */
+    discardFrame(&qf->frame);
+    if ( true(&qf->frame, FR_WATCHED) )
+    { lTop = (LocalFrame)argFrameP(&qf->frame,
+				   qf->frame.predicate->functor->arity);
+      frameFinished(&qf->frame, FINISH_CUT);
+    }
   }
 }
 
@@ -2755,26 +2762,32 @@ restore_after_query(QueryFrame qf)
 
 int
 PL_cut_query(qid_t qid)
-{ GET_LD
-  QueryFrame qf = QueryFromQid(qid);
-  int rc = TRUE;
+{ int rc = TRUE;
 
-  DEBUG(CHK_SECURE, assert(qf->magic == QID_MAGIC));
-  if ( qf->foreign_frame )
-    PL_close_foreign_frame(qf->foreign_frame);
+  if ( qid )
+  { WITH_LD(qid->engine)
+    { QueryFrame qf = QueryFromQid(qid);
 
-  if ( false(qf, PL_Q_DETERMINISTIC) )
-  { int exbefore = (exception_term != 0);
+      DEBUG(0, assert(qf->magic == QID_MAGIC));
+      if ( qf->foreign_frame )
+	PL_close_foreign_frame(qf->foreign_frame);
 
-    discard_query(qid);
-    qf = QueryFromQid(qid);
-    if ( !exbefore && exception_term != 0 )
-      rc = FALSE;
+      if ( false(qf, PL_Q_DETERMINISTIC) )
+      { int exbefore = (exception_term != 0);
+
+	discard_query(qid);
+	qf = QueryFromQid(qid);
+	if ( !exbefore && exception_term != 0 )
+	  rc = FALSE;
+      }
+
+      restore_after_query(qf);
+      qf->magic = QID_CMAGIC;		/* disqualify the frame */
+
+      freeHeap(qid, sizeof(*qid));
+    }
   }
 
-  restore_after_query(qf);
-  qf->magic = 0;			/* disqualify the frame */
-
   return rc;
 }
 
@@ -2783,28 +2796,30 @@ int
 PL_close_query(qid_t qid)
 { int rc = TRUE;
 
-  if ( qid != 0 )
-  { GET_LD
-    QueryFrame qf = QueryFromQid(qid);
+  if ( qid )
+  { WITH_LD(qid->engine)
+    { QueryFrame qf = QueryFromQid(qid);
 
-    DEBUG(CHK_SECURE, assert(qf->magic == QID_MAGIC));
-    if ( qf->foreign_frame )
-      PL_close_foreign_frame(qf->foreign_frame);
+      DEBUG(0, assert(qf->magic == QID_MAGIC));
+      if ( qf->foreign_frame )
+	PL_close_foreign_frame(qf->foreign_frame);
 
-    if ( false(qf, PL_Q_DETERMINISTIC) )
-    { int exbefore = (exception_term != 0);
+      if ( false(qf, PL_Q_DETERMINISTIC) )
+      { int exbefore = (exception_term != 0);
 
-      discard_query(qid);
-      qf = QueryFromQid(qid);
-      if ( !exbefore && exception_term != 0 )
-	rc = FALSE;
+	discard_query(qid);
+	qf = QueryFromQid(qid);
+	if ( !exbefore && exception_term != 0 )
+	  rc = FALSE;
+      }
+
+      if ( !(qf->exception && true(qf, PL_Q_PASS_EXCEPTION)) )
+	Undo(qf->choice.mark);
+
+      restore_after_query(qf);
+      qf->magic = QID_CMAGIC;		/* disqualify the frame */
+      freeHeap(qid, sizeof(*qid));
     }
-
-    if ( !(qf->exception && true(qf, PL_Q_PASS_EXCEPTION)) )
-      Undo(qf->choice.mark);
-
-    restore_after_query(qf);
-    qf->magic = 0;			/* disqualify the frame */
   }
 
   return rc;
@@ -2816,7 +2831,7 @@ PL_current_query(void)
 { GET_LD
 
   if ( HAS_LD )
-  { if ( LD->query )
+  { if ( LD->query && LD->query->magic == QID_MAGIC )
       return QidFromQuery(LD->query);
   }
 
@@ -2824,6 +2839,12 @@ PL_current_query(void)
 }
 
 
+PL_engine_t
+PL_query_engine(qid_t qid)
+{ return qid->engine;
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 PL_exception(qid) is used to extract exceptions   from an query executed
 using  PL_next_solution().  The  term-reference  itself   is  no  longer
@@ -2858,8 +2879,7 @@ PL_exception(qid_t qid)
 
 term_t
 PL_yielded(qid_t qid)
-{ GET_LD
-  QueryFrame qf = QueryFromQid(qid);
+{ QueryFrame qf = QueryFromQid(qid);
 
   return qf->yield.term;
 }
@@ -2949,6 +2969,10 @@ typedef struct register_file
 # define     __PL_ar_ctx (REGISTERS.pl_ar_ctx)
   struct foreign_context fndet_context;	/* foreign function non-deterministic context */
 # define     FNDET_CONTEXT (REGISTERS.fndet_context)
+#if VMCODE_IS_ADDRESS
+  int	     nop1;			/* See SEPARATE_VMI1 */
+  int	     nop2;			/* See SEPARATE_VMI2 */
+#endif
 #ifdef O_DEBUG
   int	     throwed_from_line;		/* Debugging: line we came from */
 # define     THROWED_FROM_LINE	(REGISTERS.throwed_from_line)
@@ -2996,7 +3020,8 @@ typedef struct register_file
 #define VMH_GOTO(...)		do { _VMH_GOTO(__VA_ARGS__); } while(0)
 #define SOLUTION_RETURN(val)	do { VMI_EXIT; _SOLUTION_RETURN(val); } while(0)
 #define VMI_GOTO_CODE(c)	do { VMI_EXIT; _VMI_GOTO_CODE(c); } while(0)
-#define SEPARATE_VMI		(void)0 /* only needed for !O_VMI_FUNCTIONS && VMCODE_IS_ADDRESS */
+#define SEPARATE_VMI1		(void)0
+#define SEPARATE_VMI2		(void)0
 
 /* By default, instruction and helper prologue/epilogue are empty */
 #define _VMI_PROLOGUE(Name,f,na,a)	;
@@ -3102,13 +3127,18 @@ static vmi_instr jmp_table[] =
 #define _NEXT_INSTRUCTION		DbgPrintInstruction(FR, PC); _VMI_GOTO_CODE(*PC++)
 #define _VMI_GOTO(n)			goto n ## _LBL
 #define _VMI_GOTO_CODE(c)		goto *(void *)(c)
-#undef SEPARATE_VMI
+#undef SEPARATE_VMI1
+#undef SEPARATE_VMI2
 /* This macro must ensure that two identical VMI instructions do not get
  * merged onto the same address by the compiler, causing decompilation
  * which translates the addresses back into the VMI number to fail.
- * initWamTable() verfies this does not happen.
+ * initWamTable() verfies this does not happen.  We have two versions
+ * to deal with jumps from to VMI to the same implementation.
  */
-#define SEPARATE_VMI { static volatile int nop = 0; (void)nop; }
+#define SEPARATE_VMI1 \
+	{ if ( ++REGISTERS.nop1 == 0 ) separate_vmi(REGISTERS.nop1); }
+#define SEPARATE_VMI2 \
+	{ if ( ++REGISTERS.nop2 == 0 ) separate_vmi(REGISTERS.nop2); }
 
 #else /* VMCODE_IS_ADDRESS */
 
@@ -3132,9 +3162,14 @@ API_STUB(int)
 
 int
 PL_next_solution(DECL_LD qid_t qid)
-{ register_file REGISTERS = {.qid = qid};	/* Active registers */
-  Code PC;					/* program counter */
-  exception_frame THROW_ENV;			/* PL_thow() environment */
+{ register_file REGISTERS =
+  { .qid = qid
+#if VMCODE_IS_ADDRESS
+    , .nop1 = 0, .nop2 = 0
+#endif
+  };					/* Active registers */
+  Code PC;				/* program counter */
+  exception_frame THROW_ENV;		/* PL_thow() environment */
 
 #if O_VMI_FUNCTIONS
   register_file *registers = &REGISTERS;
@@ -3189,6 +3224,8 @@ depart_continue() to do the normal thing or to the backtrack point.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
   QF  = QueryFromQid(qid);
+  if ( QF->magic == QID_CMAGIC )
+    return FALSE;
   DEBUG(CHK_SECURE, assert(QF->magic == QID_MAGIC));
   if ( true(QF, PL_Q_DETERMINISTIC) )	/* last one succeeded */
   { fid_t fid = QF->foreign_frame;
